//-----------------------------------------------------------------------------
// Copyright (c) 2008-2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace Dhpoware
{
    public struct TerrainVertex : IVertexType
    {
    #region Fields
        public Vector3 Position;
        public Vector2 TexCoord;
        public Vector3 Normal;
        public Vector3 Tangent;

        public static readonly VertexDeclaration VertexDeclaration = new VertexDeclaration
        (
            new VertexElement(0, VertexElementFormat.Vector3, VertexElementUsage.Position, 0),
            new VertexElement(sizeof(float) * 3, VertexElementFormat.Vector2, VertexElementUsage.TextureCoordinate, 0),
            new VertexElement(sizeof(float) * 5, VertexElementFormat.Vector3, VertexElementUsage.Normal, 0),
            new VertexElement(sizeof(float) * 8, VertexElementFormat.Vector3, VertexElementUsage.Tangent, 0)
        );
    #endregion

    #region Public methods
        public TerrainVertex(Vector3 position, Vector2 texCoord, Vector3 normal, Vector3 tangent)
        {
            Position = position;
            TexCoord = texCoord;
            Normal = normal;
            Tangent = tangent;
        }
                      
        public static void CalcTangent(ref Vector3 pos1, ref Vector3 pos2, ref Vector3 pos3,
                                       ref Vector2 texCoord1, ref Vector2 texCoord2, ref Vector2 texCoord3,
                                       ref Vector3 normal, out Vector3 tangent)
        {
            Vector3 edge1 = pos3 - pos1;
            Vector3 edge2 = pos2 - pos1;
                                                
            Vector2 texEdge1 = texCoord3 - texCoord1;
            Vector2 texEdge2 = texCoord2 - texCoord1;

            edge1.Normalize();
            edge2.Normalize();

            texEdge1.Normalize();
            texEdge2.Normalize();

            float det = (texEdge1.X * texEdge2.Y) - (texEdge1.Y * texEdge2.X);

            Vector3 t;
            Vector3 b;
            
            if ((float)Math.Abs(det) < 1e-6f)
            {
                t = Vector3.UnitX;
                b = Vector3.UnitY;
            }
            else
            {
                det = 1.0f / det;

                t.X = (texEdge2.Y * edge1.X - texEdge1.Y * edge2.X) * det;
                t.Y = (texEdge2.Y * edge1.Y - texEdge1.Y * edge2.Y) * det;
                t.Z = (texEdge2.Y * edge1.Z - texEdge1.Y * edge2.Z) * det;

                b.X = (-texEdge2.X * edge1.X + texEdge1.X * edge2.X) * det;
                b.Y = (-texEdge2.X * edge1.Y + texEdge1.X * edge2.Y) * det;
                b.Z = (-texEdge2.X * edge1.Z + texEdge1.X * edge2.Z) * det;

                t.Normalize();
                b.Normalize();
            }

            tangent.X = t.X;
            tangent.Y = t.Y;
            tangent.Z = t.Z;
        }
    #endregion

    #region Internal methods
        VertexDeclaration IVertexType.VertexDeclaration
        {
            get { return VertexDeclaration; }
        }
    #endregion
    }


    public class Terrain
    {
    #region Nested types
        public struct Region
        {
            public float Min;
            public float Max;
            public Texture2D ColorMap;
            public Texture2D NormalMap;
        }
    #endregion

    #region Fields
        private GraphicsDevice graphicsDevice;
        private ContentManager content;
        private HeightMap heightMap;
        private List<Region> regions;
        private TerrainVertex[] vertices;
        private ushort[] indices16Bit;
        private uint[] indices32Bit;
        private IndexBuffer indexBuffer;
        private DynamicVertexBuffer vertexBuffer;
        private BoundingBox boundingBox;
        private BoundingSphere boundingSphere;
        private Effect effect;
        private Vector3 sunlightDirection;
        private Vector4 sunlightColor;
        private Vector4 terrainAmbient;
        private Vector4 terrainDiffuse;
        private float terrainTilingFactor;
        private Texture2D defaultColorMap;
        private Texture2D defaultNormalMap;
        private bool enableColorMaps;
        private bool enableNormalMaps;
    #endregion

    #region Properties
        public bool EnableColorMaps
        {
            get { return enableColorMaps; }
            set { enableColorMaps = value; }
        }
        
        public bool EnableNormalMaps
        {
            get { return enableNormalMaps; }
            set { enableNormalMaps = value; }
        }

        public Vector3 SunlightDirection
        {
            get { return sunlightDirection; }
            set { sunlightDirection = value; }
        }

        public Vector4 SunlightColor
        {
            get { return sunlightColor; }
            set { sunlightColor = value; }
        }

        public Vector4 TerrainAmbient
        {
            get { return terrainAmbient; }
            set { terrainAmbient = value; }
        }

        public BoundingBox TerrainBoundingBox
        {
            get { return boundingBox; }
        }

        public BoundingSphere TerrainBoundingSphere
        {
            get { return boundingSphere; }
        }

        public Vector3 TerrainCenter
        {
            get
            {
                Vector3 center;
                float worldSize = heightMap.Size * heightMap.GridSpacing;

                center.X = worldSize * 0.5f;
                center.Z = worldSize * 0.5f;
                center.Y = heightMap.HeightAt(center.X, center.Z);

                return center;
            }
        }

        public Vector4 TerrainDiffuse
        {
            get { return terrainDiffuse; }
            set { terrainDiffuse = value; }
        }

        public HeightMap TerrainHeightMap
        {
            get { return heightMap; }
        }

        public List<Region> TerrainRegions
        {
            get { return regions; }
        }

        public float TerrainSize
        {
            get { return heightMap.Size * heightMap.GridSpacing; }
        }

        public float TerrainTilingFactor
        {
            get { return terrainTilingFactor; }
            set { terrainTilingFactor = value; }
        }
    #endregion

    #region Public methods
        public Terrain(GraphicsDevice graphicsDevice, ContentManager content)
        {
            this.graphicsDevice = graphicsDevice;
            this.content = content;

            regions = new List<Region>();

            sunlightDirection = Vector3.Down;
            sunlightColor = new Vector4(1.0f, 1.0f, 1.0f, 1.0f);
            
            terrainAmbient = new Vector4(0.3f, 0.3f, 0.3f, 1.0f);
            terrainDiffuse = new Vector4(0.7f, 0.7f, 0.7f, 1.0f);
            terrainTilingFactor = 12.0f;

            enableColorMaps = true;
            enableNormalMaps = true;
        }

        public void AddRegion(float min, float max, string colorMapAssetName, string normalMapAssetName)
        {
            // Add the new terrain region.
    
            Region region = new Region();

            region.Min = min;
            region.Max = max;
            region.ColorMap = content.Load<Texture2D>(colorMapAssetName);
            region.NormalMap = content.Load<Texture2D>(normalMapAssetName);

            regions.Add(region);

            // Calculate the new terrain tiling factor.

            float averageColorMapWidth = 0.0f;
            float averageColorMapHeight = 0.0f;

            foreach (Region r in regions)
            {
                averageColorMapWidth += r.ColorMap.Width;
                averageColorMapHeight += r.ColorMap.Height;
            }

            averageColorMapWidth /= regions.Count;
            averageColorMapHeight /= regions.Count;

            terrainTilingFactor = (float)Math.Min(averageColorMapWidth, averageColorMapHeight);
            terrainTilingFactor = TerrainSize / terrainTilingFactor;
        }

        public void Create(int size, int gridSpacing, float minHeight, float maxHeight)
        {
            heightMap = new HeightMap(size, gridSpacing, minHeight, maxHeight);
                        
            // Generate the terrain vertices.

            int totalVertices = size * size;
            int totalIndices = (size - 1) * (size - 1) * 6;

            vertices = new TerrainVertex[totalVertices];
            vertexBuffer = new DynamicVertexBuffer(graphicsDevice, typeof(TerrainVertex), vertices.Length, BufferUsage.WriteOnly);

            // Generate the terrain triangle indices.

            if (Using16BitIndices())
            {
                indices16Bit = new ushort[totalIndices];
                GenerateIndices();

                indexBuffer = new IndexBuffer(graphicsDevice, typeof(ushort), indices16Bit.Length, BufferUsage.WriteOnly);
                indexBuffer.SetData(indices16Bit);
            }
            else
            {
                indices32Bit = new uint[totalIndices];
                GenerateIndices();

                indexBuffer = new IndexBuffer(graphicsDevice, typeof(uint), indices32Bit.Length, BufferUsage.WriteOnly);
                indexBuffer.SetData(indices32Bit);
            }

            // Calculate the bounding box and bounding sphere for the terrain.

            Vector3 min, max;

            min.X = 0.0f;
            min.Y = heightMap.MinHeight;
            min.Z = 0.0f;

            max.X = heightMap.Size * heightMap.GridSpacing;
            max.Y = heightMap.MaxHeight;
            max.Z = heightMap.Size * heightMap.GridSpacing;

            boundingBox = new BoundingBox(min, max);
            boundingSphere = BoundingSphere.CreateFromBoundingBox(boundingBox);

            // Load the effect file used to render the terrain.

            try
            {
                effect = content.Load<Effect>(@"Effects\terrain");
            }
            catch (ContentLoadException)
            {
                try
                {
                    effect = content.Load<Effect>("terrain");
                }
                catch (ContentLoadException)
                {
                }
            }

            // Create the null textures.

            defaultColorMap = new Texture2D(graphicsDevice, 2, 2, false, SurfaceFormat.Color);
            Color[] whitePixels = { Color.White, Color.White, Color.White, Color.White };
            defaultColorMap.SetData(whitePixels);

            defaultNormalMap = new Texture2D(graphicsDevice, 2, 2, false, SurfaceFormat.Color);
            Color upColor = new Color(127, 127, 255);
            Color[] normalMapPixels = { upColor, upColor, upColor, upColor };
            defaultNormalMap.SetData(normalMapPixels);
        }
                
        public void Draw(Matrix world, Matrix view, Matrix projection)
        {
            if (vertexBuffer.IsContentLost)
                vertexBuffer.SetData(vertices);

            graphicsDevice.SetVertexBuffer(vertexBuffer);
            graphicsDevice.Indices = indexBuffer;

            UpdateEffect(world, view, projection);

            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Apply();

                graphicsDevice.DrawIndexedPrimitives(
                    PrimitiveType.TriangleList, 0, 0, vertices.Length, 0,
                    Using16BitIndices() ? indices16Bit.Length / 3 : indices32Bit.Length / 3);
            }

            graphicsDevice.SetVertexBuffer(null);
            graphicsDevice.Indices = null;
        }

        public void GenerateUsingDiamondSquareFractal(float roughness)
        {
            heightMap.GenerateDiamondSquareFractal(roughness);
            GenerateVertices();
        }
    #endregion

    #region Private methods
        private void GenerateIndices()
        {
            int index = 0;
            int currentVertex = 0;
            int size = heightMap.Size;

            if (Using16BitIndices())
            {
                for (int z = 0; z < size - 1; ++z)
                {
                    for (int x = 0; x < size - 1; ++x)
                    {
                        currentVertex = z * size + x;

                        indices16Bit[index++] = (ushort)(currentVertex);
                        indices16Bit[index++] = (ushort)(currentVertex + 1);
                        indices16Bit[index++] = (ushort)(currentVertex + size);

                        indices16Bit[index++] = (ushort)(currentVertex + 1);
                        indices16Bit[index++] = (ushort)(currentVertex + size + 1);
                        indices16Bit[index++] = (ushort)(currentVertex + size);
                    }
                }
            }
            else
            {
                for (int z = 0; z < size - 1; ++z)
                {
                    for (int x = 0; x < size - 1; ++x)
                    {
                        currentVertex = z * size + x;

                        indices32Bit[index++] = (uint)(currentVertex);
                        indices32Bit[index++] = (uint)(currentVertex + 1);
                        indices32Bit[index++] = (uint)(currentVertex + size);

                        indices32Bit[index++] = (uint)(currentVertex + 1);
                        indices32Bit[index++] = (uint)(currentVertex + size + 1);
                        indices32Bit[index++] = (uint)(currentVertex + size);
                    }
                }
            }
        }

        private void GenerateVertices()
        {
            int index = 0;
            int size = heightMap.Size;
            int gridSpacing = heightMap.GridSpacing;
            Vector3 position;
            Vector3 normal;
            Vector2 texCoord;

            for (int z = 0; z < size; ++z)
            {
                for (int x = 0; x < size; ++x)
                {
                    index = z * size + x;
                    
                    position.X = (float)(x * gridSpacing);
                    position.Y = heightMap.HeightAtPixel(x, z);
                    position.Z = (float)(z * gridSpacing);

                    heightMap.NormalAtPixel(x, z, out normal);

                    texCoord.X = (float)x / (float)size;
                    texCoord.Y = (float)z / (float)size;

                    vertices[index] = new TerrainVertex(position, texCoord, normal, Vector3.Zero);
                }
            }

            GenerateTangents();
            
            vertexBuffer.SetData(vertices);
        }

        private void GenerateTangents()
        {
            bool using16BitIndices = Using16BitIndices();
            int totalIndices = using16BitIndices ? indices16Bit.Length : indices32Bit.Length;
            uint[] triangle = {0, 0, 0};
            Vector3 normal = new Vector3();
            Vector3 tangent = new Vector3();

            for (int i = 0; i < totalIndices; i += 3)
            {
                if (using16BitIndices)
                {
                    triangle[0] = indices16Bit[i];
                    triangle[1] = indices16Bit[i + 1];
                    triangle[2] = indices16Bit[i + 2];
                }
                else
                {
                    triangle[0] = indices32Bit[i];
                    triangle[1] = indices32Bit[i + 1];
                    triangle[2] = indices32Bit[i + 2];
                }

                normal = Vector3.Zero;
                normal += vertices[triangle[0]].Normal;
                normal += vertices[triangle[1]].Normal;
                normal += vertices[triangle[2]].Normal;
                normal.Normalize();

                tangent = Vector3.Zero;

                TerrainVertex.CalcTangent(
                    ref vertices[triangle[0]].Position,
                    ref vertices[triangle[1]].Position,
                    ref vertices[triangle[2]].Position,
                    ref vertices[triangle[0]].TexCoord,
                    ref vertices[triangle[1]].TexCoord,
                    ref vertices[triangle[2]].TexCoord,
                    ref normal,
                    out tangent);

                vertices[triangle[0]].Tangent += tangent;
                vertices[triangle[1]].Tangent += tangent;
                vertices[triangle[2]].Tangent += tangent;
            }

            for (int i = 0; i < vertices.Length; ++i)
            {
                normal = vertices[i].Normal;
                tangent = vertices[i].Tangent;

                vertices[i].Tangent = (tangent - normal * Vector3.Dot(normal, tangent));
                vertices[i].Tangent.Normalize();
            }
        }

        private void UpdateEffect(Matrix world, Matrix view, Matrix projection)
        {
            Matrix worldInvTrans = Matrix.Transpose(Matrix.Invert(world));
            Matrix worldViewProj = world * view * projection;

            effect.Parameters["world"].SetValue(world);
            effect.Parameters["worldInvTrans"].SetValue(worldInvTrans);
            effect.Parameters["worldViewProjection"].SetValue(worldViewProj);

            effect.Parameters["sunlightDir"].SetValue(sunlightDirection);
            effect.Parameters["sunlightColor"].SetValue(sunlightColor);
            
            effect.Parameters["terrainAmbient"].SetValue(terrainAmbient);
            effect.Parameters["terrainDiffuse"].SetValue(terrainDiffuse);
            
            effect.Parameters["terrainTilingFactor"].SetValue(terrainTilingFactor);

            effect.Parameters["terrainRegion1"].SetValue(new Vector2(regions[0].Min, regions[0].Max));
            effect.Parameters["terrainRegion2"].SetValue(new Vector2(regions[1].Min, regions[1].Max));
            effect.Parameters["terrainRegion3"].SetValue(new Vector2(regions[2].Min, regions[2].Max));
            effect.Parameters["terrainRegion4"].SetValue(new Vector2(regions[3].Min, regions[3].Max));

            if (enableColorMaps)
            {
                effect.Parameters["region1ColorMapTexture"].SetValue(regions[0].ColorMap);
                effect.Parameters["region2ColorMapTexture"].SetValue(regions[1].ColorMap);
                effect.Parameters["region3ColorMapTexture"].SetValue(regions[2].ColorMap);
                effect.Parameters["region4ColorMapTexture"].SetValue(regions[3].ColorMap);
            }
            else
            {
                effect.Parameters["region1ColorMapTexture"].SetValue(defaultColorMap);
                effect.Parameters["region2ColorMapTexture"].SetValue(defaultColorMap);
                effect.Parameters["region3ColorMapTexture"].SetValue(defaultColorMap);
                effect.Parameters["region4ColorMapTexture"].SetValue(defaultColorMap);
            }

            if (enableNormalMaps)
            {
                effect.Parameters["region1NormalMapTexture"].SetValue(regions[0].NormalMap);
                effect.Parameters["region2NormalMapTexture"].SetValue(regions[1].NormalMap);
                effect.Parameters["region3NormalMapTexture"].SetValue(regions[2].NormalMap);
                effect.Parameters["region4NormalMapTexture"].SetValue(regions[3].NormalMap);
            }
            else
            {
                effect.Parameters["region1NormalMapTexture"].SetValue(defaultNormalMap);
                effect.Parameters["region2NormalMapTexture"].SetValue(defaultNormalMap);
                effect.Parameters["region3NormalMapTexture"].SetValue(defaultNormalMap);
                effect.Parameters["region4NormalMapTexture"].SetValue(defaultNormalMap);
            }
            
            effect.CurrentTechnique = effect.Techniques["TerrainTexturing"];
        }

        private bool Using16BitIndices()
        {
            return vertices.Length <= ushort.MaxValue;
        }
    #endregion
    }
}